"""
Pyfranca lexer tests.
"""

import unittest

from pyfranca import Lexer


class BaseTestCase(unittest.TestCase):

    @staticmethod
    def _tokenize(data):
        lexer = Lexer()
        tokenized_data = lexer.tokenize_data(data)
        return tokenized_data


class TestCheckRegularExpressions(BaseTestCase):
    """Test regular expression used by the lexer """

    def test_integer_valid_syntax(self):
        """test an integer """
        tokenized_data = self._tokenize("1234\n2345\n0x1234\n0X56789\n0xabcdef\n0XABCDEF\n0b10\n0B101")
        self.assertEqual(tokenized_data[0].type, "INTEGER_VAL")
        self.assertEqual(tokenized_data[0].value, 1234)

        self.assertEqual(tokenized_data[1].type, "INTEGER_VAL")
        self.assertEqual(tokenized_data[1].value, 2345)

        self.assertEqual(tokenized_data[2].type, "HEXADECIMAL_VAL")
        self.assertEqual(tokenized_data[2].value, 0x1234)

        self.assertEqual(tokenized_data[3].type, "HEXADECIMAL_VAL")
        self.assertEqual(tokenized_data[3].value, 0x56789)

        self.assertEqual(tokenized_data[4].type, "HEXADECIMAL_VAL")
        self.assertEqual(tokenized_data[4].value, 0xabcdef)

        self.assertEqual(tokenized_data[5].type, "HEXADECIMAL_VAL")
        self.assertEqual(tokenized_data[5].value, 0xabcdef)

        self.assertEqual(tokenized_data[6].type, "BINARY_VAL")
        self.assertEqual(tokenized_data[6].value, 0b10)

        self.assertEqual(tokenized_data[7].type, "BINARY_VAL")
        self.assertEqual(tokenized_data[7].value, 0b101)

    def test_string_valid_syntax(self):
        """test a string """
        tokenized_data = self._tokenize("\"This is a string\"")
        self.assertEqual(tokenized_data[0].type, "STRING_VAL")
        self.assertEqual(tokenized_data[0].value, "This is a string")

        tokenized_data = self._tokenize("\"This is a string \n with an newline\"")
        self.assertEqual(tokenized_data[0].type, "STRING_VAL")
        self.assertEqual(tokenized_data[0].value, "This is a string \n with an newline")

    def test_boolean_valid_syntax(self):
        """test a boolean value """
        tokenized_data = self._tokenize("true\nfalse")
        self.assertEqual(tokenized_data[0].type, "BOOLEAN_VAL")
        self.assertEqual(tokenized_data[0].value, True)
        self.assertEqual(tokenized_data[1].type, "BOOLEAN_VAL")
        self.assertEqual(tokenized_data[1].value, False)

    def test_integerinvalid_syntax(self):
        """test a boolean value """
        tokenized_data = self._tokenize("0xgabcdefg")
        for t in tokenized_data:
            self.assertNotEqual(t.type, "HEXADECIMAL_VAL")

    def test_booleaninvalid_syntax(self):
        """test a boolean value """
        tokenized_data = self._tokenize("istrue\nisfalse")
        for t in tokenized_data:
            self.assertNotEqual(t.type, "BOOLEAN_VAL")

    def test_float_valid_syntax(self):
        """test a float value """
        tokenized_data = self._tokenize("1.1f\n-2.2f\n3.3e3f\n-4.4e4f\n5.5e-5f"
                                        "-6.6e-6f\n0.00001f\n-0.000002f\n1e4f\n-1e4f"
                                        ".1f\n-.2f\n.3e3f\n-.4e4f\n.5e-5f"
                                        "-.6e-6f\n.00001f\n-.000002f"
                                        )
        cnt = 0
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[0].value, "1.1f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-2.2f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "3.3e3f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-4.4e4f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "5.5e-5f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-6.6e-6f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "0.00001f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-0.000002f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "1e4f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-1e4f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".1f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.2f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".3e3f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.4e4f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".5e-5f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.6e-6f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".00001f")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.000002f")

    def test_double_valid_syntax(self):
        """test a double value """
        tokenized_data = self._tokenize("1.1d\n-2.2d\n3.3e3d\n-4.4e4d\n5.5e-5d"
                                        "-6.6e-6d\n0.00001d\n-0.000002d\n1e4d\n-1e4d"
                                        ".1d\n-.2d\n.3e3d\n-.4e4d\n.5e-5d"
                                        "-.6e-6d\n.00001d\n-.000002d"
                                        )

        cnt = 0
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[0].value, "1.1d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-2.2d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "3.3e3d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-4.4e4d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "5.5e-5d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-6.6e-6d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "0.00001d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-0.000002d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "1e4d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-1e4d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".1d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.2d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".3e3d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.4e4d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".5e-5d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.6e-6d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".00001d")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.000002d")

    def test_real_valid_syntax(self):
        """test a real value """
        tokenized_data = self._tokenize("1.1\n-2.2\n3.3e3\n-4.4e4\n5.5e-5"
                                        "-6.6e-6\n0.00001\n-0.000002\n1e4\n-1e4"
                                        ".1\n-.2\n.3e3\n-.4e4\n.5e-5"
                                        "-.6e-6\n.00001\n-.000002"
                                        )

        cnt = 0
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[0].value, "1.1")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-2.2")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "3.3e3")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-4.4e4")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "5.5e-5")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-6.6e-6")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "0.00001")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-0.000002")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "1e4")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-1e4")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".1")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.2")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".3e3")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.4e4")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".5e-5")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.6e-6")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, ".00001")
        cnt += 1
        self.assertEqual(tokenized_data[cnt].type, "REAL_VAL")
        self.assertEqual(tokenized_data[cnt].value, "-.000002")

    def test_doublefloat_invalid_syntax(self):
        """test a text containing .f """
        tokenized_data = self._tokenize("""
                package org.franca.examples
               0ef .ef  -1ef   ef ed .ed
                }
            """)
        for t in tokenized_data:
            self.assertNotEqual(t.type, "REAL_VAL")

    def test_type_valid_syntax(self):
        """test an integer """
        tokenized_data = self._tokenize("const Boolean b1 = true")
        self.assertEqual(tokenized_data[0].type, "CONST")
        self.assertEqual(tokenized_data[0].value, 'const')
        self.assertEqual(tokenized_data[1].type, "BOOLEAN")
        self.assertEqual(tokenized_data[1].value, 'Boolean')
        self.assertEqual(tokenized_data[2].type, "ID")
        self.assertEqual(tokenized_data[2].value, 'b1')
        self.assertEqual(tokenized_data[3].type, "=")
        self.assertEqual(tokenized_data[3].value, '=')
        self.assertEqual(tokenized_data[4].type, "BOOLEAN_VAL")
        self.assertEqual(tokenized_data[4].value, True)

    def test_type_invalid_syntax(self):
        """test an integer """
        tokenized_data = self._tokenize("const boolean b1 = true")
        self.assertEqual(tokenized_data[0].type, "CONST")
        self.assertEqual(tokenized_data[0].value, 'const')
        self.assertEqual(tokenized_data[1].type, "ID")
        self.assertEqual(tokenized_data[1].value, 'boolean')
        self.assertEqual(tokenized_data[2].type, "ID")
        self.assertEqual(tokenized_data[2].value, 'b1')
        self.assertEqual(tokenized_data[3].type, "=")
        self.assertEqual(tokenized_data[3].value, '=')
        self.assertEqual(tokenized_data[4].type, "BOOLEAN_VAL")
        self.assertEqual(tokenized_data[4].value, True)



"""
Pyfranca lexer and parser tests.
"""

import unittest

from pyfranca import LexerException, ParserException, Parser, ast


class BaseTestCase(unittest.TestCase):

    @staticmethod
    def _parse(data):
        parser = Parser()
        package = parser.parse(data)
        return package

    def _assertParse(self, data):
        package = self._parse(data)
        self.assertIsInstance(package, ast.Package)
        return package


class TestTopLevel(BaseTestCase):
    """Test parsing top-level Franca elements."""

    def test_empty(self):
        """A package statement is expected as a minimum."""
        with self.assertRaises(ParserException) as context:
            self._parse("")
        self.assertEqual(str(context.exception),
                         "Reached unexpected end of file.")

    def test_garbage(self):
        """Invalid input."""
        with self.assertRaises(LexerException) as context:
            self._parse("%!@#")
        self.assertEqual(str(context.exception),
                         "Illegal character '%' at line 1.")

    def test_single_line_comments(self):
        self._assertParse("""
            // Package P
            package P   // This is P
            // EOF
        """)

    def test_multiline_comments(self):
        self._assertParse("""
            /* Package P
            */
            package P   /* This is P */
            /* EOF
            */
        """)

    def test_structured_comments(self):
        package = self._assertParse("""
            <** @description: Package P
                              multiline text
                @author      Jens Baumann
                @deprecated :
                @source_uri  https://github.com/zayfod/pyfranca
                @source_alias : a better franca idl parser
                @see :    foobar
                @experimental :
            **>
            package P
        """)
        self.assertEqual(package.name, "P")
        self.assertEqual(package.comments["@description"], "Package P\n                              multiline text")
        self.assertEqual(package.comments["@author"], "Jens Baumann")
        self.assertEqual(package.comments["@deprecated"], "")
        self.assertEqual(package.comments["@source_uri"], "https://github.com/zayfod/pyfranca")
        self.assertEqual(package.comments["@source_alias"], "a better franca idl parser")
        self.assertEqual(package.comments["@see"], "foobar")
        self.assertEqual(package.comments["@experimental"], "")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 0)
        self.assertEqual(len(package.interfaces), 0)

    def test_structured_comments_complete_example(self):
        package = self._assertParse("""
                   package P

                   <** @description : collection TC   **>
                   typeCollection TC {

                       <** @description : struct S   **>
                       struct S {

                           <** @description : member a   **>
                           Int32 a

                           <** @description : member b   **>
                           String b
                       }

                       <** @description : enumeration E   **>
                       enumeration E {
                          <** @description : enum member a   **>
                          a = 0x1
                          <** @description : enum member b   **>
                          b = 0x2
                       }

                       <** @description : typedef  MyInt8  **>
                       typedef MyInt8 is Int8

                       <** @description : array A  **>
                       array A of UInt8

                        <** @description : map M  **>
                        map M {
                            String to Int32
                        }
                    }
                    <** @description : interface I **>
                    interface I {
                        version { major 1 minor 0 }

                        <** @description : attribute A **>
                        attribute Int32 A

                        <** @description : method M **>
                        method M {
                            in {
                                <** @description : in parameter a **>
                                Float a

                                <** @description : in parameter b **>
                                Float[] b
                            }
                            out {
                                <** @description : out parameter result **>
                                Float result
                            }
                            error {
                                <** @description : error FAILURE **>
                                FAILURE = 0

                                <** @description : error SUCCESS **>
                                SUCCESS
                            }
                        }

                        <** @description : broadcast B **>
                        broadcast B {
                            out {
                                <** @description : out parameter x **>
                                Int32 x
                            }
                        }
                    }
               """)
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 1)
        self.assertEqual(len(package.interfaces), 1)
        self.assertEqual(package.typecollections['TC'].comments["@description"], "collection TC")
        self.assertEqual(package.typecollections['TC'].structs['S'].comments["@description"], "struct S")
        self.assertEqual(package.typecollections['TC'].structs['S'].fields['a'].comments["@description"], "member a")
        self.assertEqual(package.typecollections['TC'].structs['S'].fields['b'].comments["@description"], "member b")
        self.assertEqual(package.typecollections['TC'].enumerations['E'].comments['@description'], "enumeration E")
        self.assertEqual(package.typecollections['TC'].enumerations['E'].enumerators['a'].value.value, 1)
        self.assertEqual(
            package.typecollections['TC'].enumerations['E'].enumerators['a'].comments["@description"],
            "enum member a")
        self.assertEqual(package.typecollections['TC'].enumerations['E'].enumerators['b'].value.value, 2)
        self.assertEqual(
            package.typecollections['TC'].enumerations['E'].enumerators['b'].comments['@description'],
            "enum member b")

        self.assertEqual(package.interfaces['I'].comments['@description'], "interface I")
        self.assertEqual(package.interfaces['I'].attributes['A'].comments['@description'], "attribute A")

        self.assertEqual(package.interfaces['I'].methods['M'].comments['@description'], "method M")
        self.assertEqual(package.interfaces['I'].methods['M'].in_args['a'].comments['@description'], "in parameter a")
        self.assertEqual(package.interfaces['I'].methods['M'].in_args['b'].comments['@description'], "in parameter b")
        self.assertEqual(package.interfaces['I'].methods['M'].out_args['result'].comments['@description'],
                         "out parameter result")
        self.assertEqual(package.interfaces['I'].methods['M'].errors['FAILURE'].comments['@description'],
                         "error FAILURE")
        self.assertEqual(package.interfaces['I'].methods['M'].errors['SUCCESS'].comments['@description'],
                         "error SUCCESS")
        self.assertEqual(package.interfaces['I'].broadcasts['B'].comments['@description'], "broadcast B")
        self.assertEqual(package.interfaces['I'].broadcasts['B'].out_args['x'].comments['@description'],
                         "out parameter x")

    def test_empty_comments(self):
        package = self._assertParse("""
                    <**  **>
                    package P
                """)
        self.assertEqual(package.name, "P")
        self.assertEqual(len(package.comments.keys()), 0)

    def test_comments_no_tag(self):
        package = self._assertParse("""
                    <** no tag  **>
                    package P
                """)
        self.assertEqual(package.name, "P")
        self.assertEqual(len(package.comments.keys()), 0)

    def test_bad_comments(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
            package P

             <** @description : bad comment**>

             <** @description : typeCollection TC **>
             typeCollection TC {}

        """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 6 near '@description : typeCollection TC'.")

    def test_empty_package(self):
        package = self._assertParse("package P")
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 0)
        self.assertEqual(len(package.interfaces), 0)

    def test_multiple_package_definitions(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                package P2
            """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 3 near 'package'.")

    def test_import_namespace(self):
        package = self._assertParse("package P import NS from \"test.fidl\"")
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 1)
        self.assertEqual(len(package.typecollections), 0)
        self.assertEqual(len(package.interfaces), 0)
        imp = package.imports[0]
        self.assertEqual(imp.namespace, "NS")
        self.assertEqual(imp.file, "test.fidl")

    def test_import_model(self):
        package = self._assertParse("package P import model \"test.fidl\"")
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 1)
        self.assertEqual(len(package.typecollections), 0)
        self.assertEqual(len(package.interfaces), 0)
        imp = package.imports[0]
        self.assertIsNone(imp.namespace)
        self.assertEqual(imp.file, "test.fidl")

    def test_import_two_namespaces(self):
        package = self._assertParse("""
            package P
            import NS1 from "test1.fidl"
            import NS2 from "test2.fidl"
        """)
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 2)
        self.assertEqual(len(package.typecollections), 0)
        self.assertEqual(len(package.interfaces), 0)
        imp = package.imports[0]
        self.assertEqual(imp.namespace, "NS1")
        self.assertEqual(imp.file, "test1.fidl")
        imp = package.imports[1]
        self.assertEqual(imp.namespace, "NS2")
        self.assertEqual(imp.file, "test2.fidl")

    def test_empty_typecollection(self):
        package = self._assertParse("package P typeCollection TC {}")
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 1)
        self.assertEqual(len(package.interfaces), 0)
        self.assertTrue("TC" in package.typecollections)
        typecollection = package.typecollections["TC"]
        self.assertEqual(typecollection.package, package)
        self.assertEqual(typecollection.name, "TC")
        self.assertListEqual(typecollection.flags, [])
        self.assertIsNone(typecollection.version)
        self.assertEqual(len(typecollection.typedefs), 0)
        self.assertEqual(len(typecollection.enumerations), 0)
        self.assertEqual(len(typecollection.structs), 0)
        self.assertEqual(len(typecollection.arrays), 0)
        self.assertEqual(len(typecollection.maps), 0)

    def test_two_empty_typecollections(self):
        package = self._assertParse("""
            package P
            typeCollection TC1 {}
            typeCollection TC2 {}
        """)
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 2)
        self.assertEqual(len(package.interfaces), 0)
        self.assertTrue("TC1" in package.typecollections)
        typecollection = package.typecollections["TC1"]
        self.assertEqual(typecollection.package, package)
        self.assertEqual(typecollection.name, "TC1")
        self.assertListEqual(typecollection.flags, [])
        self.assertIsNone(typecollection.version)
        self.assertEqual(len(typecollection.typedefs), 0)
        self.assertEqual(len(typecollection.enumerations), 0)
        self.assertEqual(len(typecollection.structs), 0)
        self.assertEqual(len(typecollection.arrays), 0)
        self.assertEqual(len(typecollection.maps), 0)
        self.assertTrue("TC2" in package.typecollections)
        typecollection = package.typecollections["TC2"]
        self.assertEqual(typecollection.package, package)
        self.assertEqual(typecollection.name, "TC2")
        self.assertListEqual(typecollection.flags, [])
        self.assertIsNone(typecollection.version)
        self.assertEqual(len(typecollection.typedefs), 0)
        self.assertEqual(len(typecollection.enumerations), 0)
        self.assertEqual(len(typecollection.structs), 0)
        self.assertEqual(len(typecollection.arrays), 0)
        self.assertEqual(len(typecollection.maps), 0)

    def test_empty_interface(self):
        package = self._assertParse("package P interface I {}")
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 0)
        self.assertEqual(len(package.interfaces), 1)
        self.assertTrue("I" in package.interfaces)
        interface = package.interfaces["I"]
        self.assertEqual(interface.package, package)
        self.assertEqual(interface.name, "I")
        self.assertListEqual(interface.flags, [])
        self.assertIsNone(interface.version)
        self.assertEqual(len(interface.typedefs), 0)
        self.assertEqual(len(interface.enumerations), 0)
        self.assertEqual(len(interface.structs), 0)
        self.assertEqual(len(interface.arrays), 0)
        self.assertEqual(len(interface.maps), 0)
        self.assertEqual(len(interface.attributes), 0)
        self.assertEqual(len(interface.methods), 0)
        self.assertEqual(len(interface.broadcasts), 0)

    def test_two_empty_interfaces(self):
        package = self._assertParse("""
            package P
            interface I1 {}
            interface I2 {}
        """)
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 0)
        self.assertEqual(len(package.interfaces), 2)
        self.assertTrue("I1" in package.interfaces)
        interface = package.interfaces["I1"]
        self.assertEqual(interface.package, package)
        self.assertEqual(interface.name, "I1")
        self.assertListEqual(interface.flags, [])
        self.assertIsNone(interface.version)
        self.assertEqual(len(interface.typedefs), 0)
        self.assertEqual(len(interface.enumerations), 0)
        self.assertEqual(len(interface.structs), 0)
        self.assertEqual(len(interface.arrays), 0)
        self.assertEqual(len(interface.maps), 0)
        self.assertEqual(len(interface.attributes), 0)
        self.assertEqual(len(interface.methods), 0)
        self.assertEqual(len(interface.broadcasts), 0)
        self.assertTrue("I2" in package.interfaces)
        interface = package.interfaces["I2"]
        self.assertEqual(interface.package, package)
        self.assertEqual(interface.name, "I2")
        self.assertListEqual(interface.flags, [])
        self.assertIsNone(interface.version)
        self.assertEqual(len(interface.typedefs), 0)
        self.assertEqual(len(interface.enumerations), 0)
        self.assertEqual(len(interface.structs), 0)
        self.assertEqual(len(interface.arrays), 0)
        self.assertEqual(len(interface.maps), 0)
        self.assertEqual(len(interface.attributes), 0)
        self.assertEqual(len(interface.methods), 0)
        self.assertEqual(len(interface.broadcasts), 0)


class TestUnsupported(BaseTestCase):
    """Test that unsupported Franca features fail appropriately."""

    def test_dots_in_namespace_names(self):
        """Franca 0.9.2, section 5.8.2"""
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package org.franca.examples
                interface simple.ExampleInterface {
                    // this interface can be globally accessed by the FQN
                    // org.franca.examples.simple.ExampleInterface
                }
            """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 3 near '.'.")

    def test_expressions(self):
        """Franca 0.9.2, section 5.2.1"""
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                typeCollection TC {
                    const UInt32 twentyFive = 55
                    const Boolean b2 = MAX COUNT > 3
                    const Boolean b3 = (a && b) jj foo=="bar"
                }
            """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 5 near 'MAX'.")

    def test_complex_constants(self):
        """Franca 0.9.2, section 5.2.2"""
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                typeCollection TC {
                    array Array1 of UInt16
                    const Array1 empty = []
                    const Array1 full = [ 1, 2, 2+3, 100*100+100 ]

                    struct Struct1 {
                        Boolean e1
                        UInt16 e2
                        String e3
                    }
                    const Struct1 s1 = { e1: true, e2: 1, e3: "foo" }

                    union Union1 {
                        UInt16 e1
                        Boolean e2
                        String e3
                    }
                    const Union1 uni1 = { e1: 1 }
                    const Union1 uni2 = { e3: "foo" }

                    map Map1 { UInt16 to String }
                    const Map1 m1 = [ 1 => "one", 2 => "two" ]
                }
            """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 5 near 'Array1'.")

    def test_unions(self):
        """Franca 0.9.2, section 5.1.6"""
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                typeCollection TC {
                    union ExampleUnion {
                        UInt32 element1
                        Float element2
                    }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 4 near 'union'.")

    def test_error_extending(self):
        """Franca 0.9.2, section 5.5.3"""
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                interface I {
                    method m {
                        error extends GenericErrors {
                            OVERFLOW
                            UNDERFLOW
                        }
                    }
                    enumeration GenericErrors {
                        INVALID_PARAMATERS
                    }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 5 near 'extends'.")

    def test_contracts(self):
        """Franca 0.9.2, section 5.5.3"""
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                contract {
                    PSM {
                        initial idle
                        state idle {
                            on call setActivePlayer -> working
                        }
                        state working {
                            on signal attachOutput -> idle
                        }
                    }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 3 near 'contract'.")


class TestFrancaUserManualExamples(BaseTestCase):
    """Test parsing various examples from the Franca user manual."""

    def test_calculator_api(self):
        self._assertParse("""
            package P
            interface CalculatorAPI {
                method add {
                    in {
                        Float a
                        Float b
                    }
                    out {
                        Float sum
                    }
                }
            }
        """)

    def test_example_interface(self):
        self._assertParse("""
            package P
            interface ExampleInterface {
                attribute Double temperature readonly noSubscriptions
                attribute Boolean overheated
            }
        """)

    def test_calculator_divide(self):
        """Method "error extends ..." statements are not supported."""
        with self.assertRaises(ParserException) as context:
            self._parse("""
            package P
            interface Calculator {
                method divide {
                    in {
                        UInt32 dividend
                        UInt32 divisor
                    }
                    out {
                        UInt32 quotient
                        UInt32 remainder
                    }
                    error extends GenericErrors {
                        DIVISION_BY_ZERO
                        OVERFLOW
                        UNDERFLOW
                    }
                }
                enumeration GenericErrors {
                    INVALID_PARAMATERS
                    // ...
                }
            }
        """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 13 near 'extends'.")

    def test_calculator_divide2(self):
        self._assertParse("""
            package P
            interface Calculator {
                method divide {
                    in {
                        UInt32 dividend
                        UInt32 divisor
                    }
                    out {
                        UInt32 quotient
                        UInt32 remainder
                    }
                    error CalcErrors
                }
                enumeration CalcErrors {
                    DIVISION_BY_ZERO
                    OVERFLOW
                    UNDERFLOW
                }
            }
        """)

    def test_broadcast(self):
        self._assertParse("""
            package P
            interface ExampleInterface {
                broadcast buttonClicked {
                    out {
                        ButtonId id
                        Boolean isLongPress
                    }
                }
            }
        """)


class TestMisc(BaseTestCase):
    """Test parsing various FIDL examples."""

    def test_all_supported(self):
        self._assertParse("""
            package P

            import m1.* from "m1.fidl"
            import model "m2.fidl"

            typeCollection TC {
                version { major 1 minor 0 }

                typedef MyInt8 is Int8
                typedef MyInt16 is Int16
                typedef MyInt32 is Int32
                typedef MyInt64 is Int64
                typedef MyUInt8 is UInt8
                typedef MyUInt16 is UInt16
                typedef MyUInt32 is UInt32
                typedef MyUInt64 is UInt64
                typedef MyBoolean is Boolean
                typedef MyFloat is Float
                typedef MyDouble is Double
                typedef MyString is String
                typedef MyByteBuffer is ByteBuffer

                typedef MyInt32Array is Int32[]
                typedef MyThis is MyThat

                enumeration E {
                    FALSE = 0
                    TRUE
                }
                enumeration E2 extends E {}

                struct S {
                    Int32 a
                    Int32[] b
                }
                struct S2 extends S {}
                struct S3 polymorphic {}

                array A of UInt8

                map M {
                    String to Int32
                }
                map M2 {
                    Key to Value
                }
            }

            interface I {
                version { major 1 minor 0 }

                attribute Int32 A
                attribute Int32 A2 readonly
                attribute Int32 A3 noSubscriptions

                method M {
                    in {
                        Float a
                        Float[] b
                    }
                    out {
                        Float result
                    }
                    error {
                        FAILURE = 0
                        SUCCESS
                    }
                }
                method M2 fireAndForget {}

                broadcast B {
                    out {
                        Int32 x
                    }
                }
                broadcast B2 selective {}

                typedef ITD is String

                enumeration IE {
                    FALSE = 0
                    TRUE
                }

                struct IS {
                    Int32 a
                    Int32[] b
                }

                array IA of UInt8

                map IM {
                    String to Int32
                }
            }

            interface I2 extends I {
            }
        """)


class TestTypeCollections(BaseTestCase):
    """Test parsing type collections."""

    def test_normal(self):
        package = self._assertParse("""
            package P
            typeCollection TC {
                version { major 12 minor 34 }
            }
        """)
        tc = package.typecollections["TC"]
        self.assertEqual(tc.package, package)
        self.assertEqual(tc.name, "TC")
        self.assertEqual(tc.flags, [])
        self.assertEqual(str(tc.version), "12.34")
        self.assertEqual(len(tc.typedefs), 0)
        self.assertEqual(len(tc.enumerations), 0)
        self.assertEqual(len(tc.structs), 0)
        self.assertEqual(len(tc.arrays), 0)
        self.assertEqual(len(tc.maps), 0)

    def test_multiple_versions(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                typeCollection TC {
                    version { major 1 minor 0 }
                    version { major 2 minor 0 }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Multiple version definitions.")

    def test_duplicate_member(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                typeCollection TC {
                    typedef X is MyThat
                    array X of UInt8
                }
            """)
        self.assertEqual(str(context.exception),
                         "Duplicate namespace member 'X'.")


class TestInterfaces(BaseTestCase):
    """Test parsing interfaces."""

    def test_normal(self):
        package = self._assertParse("""
            package P
            interface I {
                version { major 12 minor 34 }
                method add {
                    in {
                        Float a
                        Float b
                    }
                    out {
                        Float sum
                    }
                }
            }
            interface I2 extends I {
            }
        """)
        i = package.interfaces["I"]
        self.assertEqual(i.package, package)
        self.assertEqual(i.name, "I")
        self.assertEqual(i.flags, [])
        self.assertEqual(str(i.version), "12.34")
        self.assertEqual(len(i.typedefs), 0)
        self.assertEqual(len(i.enumerations), 0)
        self.assertEqual(len(i.structs), 0)
        self.assertEqual(len(i.arrays), 0)
        self.assertEqual(len(i.maps), 0)
        self.assertEqual(len(i.attributes), 0)
        self.assertEqual(len(i.methods), 1)
        self.assertEqual(len(i.broadcasts), 0)
        self.assertIsNone(i.extends)
        i2 = package.interfaces["I2"]
        self.assertEqual(i2.package, package)
        self.assertEqual(i2.name, "I2")
        self.assertIsNone(i2.version)
        self.assertEqual(i2.extends, "I")

    def test_multiple_versions(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                interface I {
                    version { major 1 minor 0 }
                    version { major 2 minor 0 }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Multiple version definitions.")

    def test_duplicate_member(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                interface I {
                    typedef X is MyThat
                    method X {}
                }
            """)
        self.assertEqual(str(context.exception),
                         "Duplicate namespace member 'X'.")


class TestEnumerations(BaseTestCase):
    """Test parsing enumerations."""

    def test_normal(self):
        package = self._assertParse("""
            package P
            typeCollection TC {
                enumeration E {
                    FALSE = 0
                    TRUE
                }
                enumeration E2 extends E {}
            }
        """)
        typecollection = package.typecollections["TC"]
        e = typecollection.enumerations["E"]
        self.assertEqual(e.namespace, typecollection)
        self.assertEqual(e.name, "E")
        self.assertIsNone(e.extends)
        self.assertEqual(e.flags, [])
        self.assertEqual(len(e.enumerators), 2)
        ee = e.enumerators["FALSE"]
        self.assertEqual(ee.name, "FALSE")
        self.assertEqual(ee.value.value, 0)
        ee = e.enumerators["TRUE"]
        self.assertEqual(ee.name, "TRUE")
        self.assertIsNone(ee.value)
        e2 = typecollection.enumerations["E2"]
        self.assertEqual(e2.namespace, typecollection)
        self.assertEqual(e2.name, "E2")
        self.assertEqual(e2.extends, "E")
        self.assertEqual(e2.flags, [])
        self.assertEqual(len(e2.enumerators), 0)

    def test_hexvalue(self):
        package = self._assertParse("""
            package P
            typeCollection TC {
                enumeration E {
                    A = 0xA
                    B = 0XABC
                }
                enumeration E2 extends E {}
            }
        """)
        typecollection = package.typecollections["TC"]
        e = typecollection.enumerations["E"]
        self.assertEqual(e.namespace, typecollection)
        self.assertEqual(e.name, "E")
        self.assertIsNone(e.extends)
        self.assertEqual(e.flags, [])
        self.assertEqual(len(e.enumerators), 2)
        ee = e.enumerators["A"]
        self.assertEqual(ee.name, "A")
        self.assertEqual(ee.value.value, 10)
        ee = e.enumerators["B"]
        self.assertEqual(ee.name, "B")
        self.assertEqual(ee.value.value, 0xABC)
        e2 = typecollection.enumerations["E2"]
        self.assertEqual(e2.namespace, typecollection)
        self.assertEqual(e2.name, "E2")
        self.assertEqual(e2.extends, "E")
        self.assertEqual(e2.flags, [])
        self.assertEqual(len(e2.enumerators), 0)

    def test_binvalue(self):
        package = self._assertParse("""
            package P
            typeCollection TC {
                enumeration E {
                    A = 0b10
                    B = 0B011
                }
                enumeration E2 extends E {}
            }
        """)
        typecollection = package.typecollections["TC"]
        e = typecollection.enumerations["E"]
        self.assertEqual(e.namespace, typecollection)
        self.assertEqual(e.name, "E")
        self.assertIsNone(e.extends)
        self.assertEqual(e.flags, [])
        self.assertEqual(len(e.enumerators), 2)
        ee = e.enumerators["A"]
        self.assertEqual(ee.name, "A")
        self.assertEqual(ee.value.value, 0b10)
        ee = e.enumerators["B"]
        self.assertEqual(ee.name, "B")
        self.assertEqual(ee.value.value, 0b011)
        e2 = typecollection.enumerations["E2"]
        self.assertEqual(e2.namespace, typecollection)
        self.assertEqual(e2.name, "E2")
        self.assertEqual(e2.extends, "E")
        self.assertEqual(e2.flags, [])
        self.assertEqual(len(e2.enumerators), 0)

    def test_badvalue(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
            package P
            typeCollection TC {
                enumeration E {
                    A = 0.123f
                    B = 0xABC
                }
                enumeration E2 extends E {}
            }
        """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 5 near '0.123f'.")

    def test_duplicate_enumerator(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                typeCollection TC {
                    enumeration E { a a }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Duplicate enumerator 'a'.")


class TestStructs(BaseTestCase):
    """Test parsing structures."""

    def test_normal(self):
        package = self._assertParse("""
            package P
            typeCollection TC {
                struct S {
                    Int32 a
                    String b
                }
                struct S2 extends S {}
                struct S3 polymorphic {}
            }
            interface I {
                struct S4 {
                    Int32 a
                    String b
                }
            }
        """)
        typecollection = package.typecollections["TC"]
        s = typecollection.structs["S"]
        self.assertEqual(s.name, "S")
        self.assertIsNone(s.extends)
        self.assertEqual(s.namespace, typecollection)
        self.assertEqual(len(s.fields), 2)
        self.assertEqual(s.fields["a"].name, "a")
        self.assertIsInstance(s.fields["a"].type, ast.Int32)
        self.assertEqual(s.fields["b"].name, "b")
        self.assertIsInstance(s.fields["b"].type, ast.String)
        self.assertEqual(len(s.flags), 0)
        s2 = typecollection.structs["S2"]
        self.assertEqual(s2.name, "S2")
        self.assertEqual(s2.extends, "S")
        self.assertEqual(len(s2.fields), 0)
        self.assertEqual(s2.namespace, typecollection)
        self.assertEqual(len(s2.flags), 0)
        s3 = typecollection.structs["S3"]
        self.assertEqual(s3.name, "S3")
        self.assertIsNone(s3.extends)
        self.assertEqual(s3.namespace, typecollection)
        self.assertEqual(len(s3.fields), 0)
        self.assertEqual(len(s3.flags), 1)
        self.assertEqual(s3.flags[0], "polymorphic")
        interface = package.interfaces["I"]
        self.assertEqual(len(interface.structs), 1)
        s4 = interface.structs["S4"]
        self.assertEqual(s4.name, "S4")
        self.assertIsNone(s4.extends)
        self.assertEqual(s4.namespace, interface)
        self.assertEqual(len(s4.fields), 2)
        self.assertEqual(s4.fields["a"].name, "a")
        self.assertIsInstance(s4.fields["a"].type, ast.Int32)
        self.assertEqual(s4.fields["b"].name, "b")
        self.assertIsInstance(s4.fields["b"].type, ast.String)
        self.assertEqual(len(s4.flags), 0)

    def test_duplicate_field(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                typeCollection TC {
                    struct S {
                        Int32 a
                        String a
                    }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Duplicate structure field 'a'.")


class TestMethods(BaseTestCase):
    """Test parsing methods."""

    def test_duplicate_argument(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                interface I {
                    method M {
                        in {
                            Float a
                            Float a
                        }
                    }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Duplicate argument 'a'.")


class TestBroadcasts(BaseTestCase):
    """Test parsing broadcasts."""

    def test_duplicate_argument(self):
        with self.assertRaises(ParserException) as context:
            self._parse("""
                package P
                interface I {
                    broadcast B {
                        out {
                            Float a
                            Float a
                        }
                    }
                }
            """)
        self.assertEqual(str(context.exception), "Duplicate argument 'a'.")


class TestConstants(BaseTestCase):
    """Test parsing constants."""
    def test_constants(self):
        """Franca 0.9.2, section 5.2.1"""
        package = self._parse("""
            package P
            typeCollection TC {
                const UInt32 MAX_COUNT = 10000
                const UInt32 MAX_COUNT_HEX = 0x10000
                const UInt32 MAX_COUNT_BIN = 0b10000
                const Double pi = 3.1415d
                const Float f1 = 1.2f
                const Float f2 = 6.022e23f
                const Boolean b1 = true
                const Boolean b2 = false

                const String s1 = "Hello"

                const String s2 = "Hello
                                   World"
            }
        """)
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 1)
        self.assertEqual(len(package.interfaces), 0)
        self.assertTrue("TC" in package.typecollections)
        typecollection = package.typecollections["TC"]
        self.assertEqual(typecollection.package, package)
        self.assertEqual(typecollection.name, "TC")
        self.assertListEqual(typecollection.flags, [])
        self.assertIsNone(typecollection.version)
        self.assertEqual(len(typecollection.typedefs), 0)
        self.assertEqual(len(typecollection.enumerations), 0)
        self.assertEqual(len(typecollection.structs), 0)
        self.assertEqual(len(typecollection.arrays), 0)
        self.assertEqual(len(typecollection.maps), 0)
        self.assertEqual(len(typecollection.constants), 10)

        self.assertEqual(typecollection.constants["MAX_COUNT"].name, "MAX_COUNT")
        self.assertEqual(typecollection.constants["MAX_COUNT"].type.name, "UInt32")
        self.assertEqual(typecollection.constants["MAX_COUNT"].value.value, 10000)
        self.assertEqual(typecollection.constants["MAX_COUNT"].value.name, "IntegerValue")
        self.assertEqual(typecollection.constants["MAX_COUNT"].value.base, ast.IntegerValue.DECIMAL)

        self.assertEqual(typecollection.constants["MAX_COUNT_HEX"].name, "MAX_COUNT_HEX")
        self.assertEqual(typecollection.constants["MAX_COUNT_HEX"].type.name, "UInt32")
        self.assertEqual(typecollection.constants["MAX_COUNT_HEX"].value.value, 0x10000)
        self.assertEqual(typecollection.constants["MAX_COUNT_HEX"].value.name, "IntegerValue")
        self.assertEqual(typecollection.constants["MAX_COUNT_HEX"].value.base, ast.IntegerValue.HEXADECIMAL)

        self.assertEqual(typecollection.constants["MAX_COUNT_BIN"].name, "MAX_COUNT_BIN")
        self.assertEqual(typecollection.constants["MAX_COUNT_BIN"].type.name, "UInt32")
        self.assertEqual(typecollection.constants["MAX_COUNT_BIN"].value.value, 0b10000)
        self.assertEqual(typecollection.constants["MAX_COUNT_BIN"].value.name, "IntegerValue")
        self.assertEqual(typecollection.constants["MAX_COUNT_BIN"].value.base, ast.IntegerValue.BINARY)

        self.assertEqual(typecollection.constants["pi"].name, "pi")
        self.assertEqual(typecollection.constants["pi"].type.name, "Double")
        self.assertEqual(typecollection.constants["pi"].value.value, 3.1415)
        self.assertEqual(typecollection.constants["pi"].value.name, "DoubleValue")

        self.assertEqual(typecollection.constants["f1"].name, "f1")
        self.assertEqual(typecollection.constants["f1"].type.name, "Float")
        self.assertAlmostEqual(typecollection.constants["f1"].value.value, 1.2)
        self.assertEqual(typecollection.constants["f1"].value.name, "FloatValue")

        self.assertEqual(typecollection.constants["f2"].name, "f2")
        self.assertEqual(typecollection.constants["f2"].type.name, "Float")
        self.assertAlmostEqual(typecollection.constants["f2"].value.value, 6.022e23)
        self.assertEqual(typecollection.constants["f2"].value.name, "FloatValue")

        self.assertEqual(typecollection.constants["b1"].name, "b1")
        self.assertEqual(typecollection.constants["b1"].type.name, "Boolean")
        self.assertEqual(typecollection.constants["b1"].value.value, True)
        self.assertEqual(typecollection.constants["b1"].value.name, "BooleanValue")

        self.assertEqual(typecollection.constants["b2"].name, "b2")
        self.assertEqual(typecollection.constants["b2"].type.name, "Boolean")
        self.assertEqual(typecollection.constants["b2"].value.value, False)
        self.assertEqual(typecollection.constants["b2"].value.name, "BooleanValue")

        self.assertEqual(typecollection.constants["s1"].name, "s1")
        self.assertEqual(typecollection.constants["s1"].type.name, "String")
        self.assertAlmostEqual(typecollection.constants["s1"].value.value, "Hello")
        self.assertEqual(typecollection.constants["s1"].value.name, "StringValue")

        self.assertEqual(typecollection.constants["s2"].name, "s2")
        self.assertEqual(typecollection.constants["s2"].type.name, "String")
        self.assertAlmostEqual(typecollection.constants["s2"].value.value,
                               "Hello\n                                   World")
        self.assertEqual(typecollection.constants["s2"].value.name, "StringValue")

    def test_constants_casting(self):
        """Franca 0.9.2, section 5.2.1"""
        package = self._parse("""
            package P
            typeCollection TC {
                const Float f1 = 123
                const Float f2 = true
                const Double d1 = 123
                const Double d2 = true
                const Boolean b1 = 123
                const Boolean b2 = 123.0f
                const Boolean b3 = 0.0d
                const Boolean b4 = "123"
                const String s1 = 123
                const String s2 = 123.0f
                const String s3 = 0.0d
                const String s4 = true
            }
        """)
        self.assertEqual(package.name, "P")
        self.assertEqual(package.files, [])
        self.assertEqual(len(package.imports), 0)
        self.assertEqual(len(package.typecollections), 1)
        self.assertEqual(len(package.interfaces), 0)
        self.assertTrue("TC" in package.typecollections)
        typecollection = package.typecollections["TC"]
        self.assertEqual(typecollection.package, package)
        self.assertEqual(typecollection.name, "TC")
        self.assertListEqual(typecollection.flags, [])
        self.assertIsNone(typecollection.version)
        self.assertEqual(len(typecollection.typedefs), 0)
        self.assertEqual(len(typecollection.enumerations), 0)
        self.assertEqual(len(typecollection.structs), 0)
        self.assertEqual(len(typecollection.arrays), 0)
        self.assertEqual(len(typecollection.maps), 0)
        self.assertEqual(len(typecollection.constants), 12)

        x = "f1"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "Float")
        self.assertAlmostEqual(typecollection.constants[x].value.value, 123.0)
        self.assertEqual(typecollection.constants[x].value.name, "FloatValue")

        x = "f2"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "Float")
        self.assertAlmostEqual(typecollection.constants[x].value.value, 1.0)
        self.assertEqual(typecollection.constants[x].value.name, "FloatValue")

        x = "d1"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "Double")
        self.assertAlmostEqual(typecollection.constants[x].value.value, 123.0)
        self.assertEqual(typecollection.constants[x].value.name, "DoubleValue")

        x = "d2"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "Double")
        self.assertAlmostEqual(typecollection.constants[x].value.value, 1.0)
        self.assertEqual(typecollection.constants[x].value.name, "DoubleValue")

        x = "b1"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "Boolean")
        self.assertAlmostEqual(typecollection.constants[x].value.value, True)
        self.assertEqual(typecollection.constants[x].value.name, "BooleanValue")

        x = "b2"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "Boolean")
        self.assertAlmostEqual(typecollection.constants[x].value.value, True)
        self.assertEqual(typecollection.constants[x].value.name, "BooleanValue")

        x = "b3"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "Boolean")
        self.assertAlmostEqual(typecollection.constants[x].value.value, False)
        self.assertEqual(typecollection.constants[x].value.name, "BooleanValue")

        x = "b4"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "Boolean")
        self.assertAlmostEqual(typecollection.constants[x].value.value, True)
        self.assertEqual(typecollection.constants[x].value.name, "BooleanValue")

        x = "s1"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "String")
        self.assertAlmostEqual(typecollection.constants[x].value.value, "123")
        self.assertEqual(typecollection.constants[x].value.name, "StringValue")

        x = "s2"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "String")
        self.assertAlmostEqual(typecollection.constants[x].value.value, "123.0")
        self.assertEqual(typecollection.constants[x].value.name, "StringValue")

        x = "s3"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "String")
        self.assertAlmostEqual(typecollection.constants[x].value.value, "0.0")
        self.assertEqual(typecollection.constants[x].value.name, "StringValue")

        x = "s4"
        self.assertEqual(typecollection.constants[x].name, x)
        self.assertEqual(typecollection.constants[x].type.name, "String")
        self.assertAlmostEqual(typecollection.constants[x].value.value, "True")
        self.assertEqual(typecollection.constants[x].value.name, "StringValue")

    def test_constants_bad_syntax_Uint32(self):
        """Franca 0.9.2, section 5.2.1"""

        with self.assertRaises(ParserException) as context:
            self._parse("""
            package P
            typeCollection TC {
                const UInt32 MAX_COUNT = "Hello"
            }
        """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 4 near 'Hello'.")

    def test_constants_bad_syntax_String(self):
        """Franca 0.9.2, section 5.2.1"""

        with self.assertRaises(ParserException) as context:
            self._parse("""
            package P
            typeCollection TC {
                const String s1 = 123abc
            }
        """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 4 near 'abc'.")

    def test_constants_bad_syntax_Boolean(self):
        """Franca 0.9.2, section 5.2.1"""

        with self.assertRaises(ParserException) as context:
            self._parse("""
            package P
            typeCollection TC {
                const Boolean b1 = 123asc
            }
        """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 4 near 'asc'.")

    def test_constants_bad_syntax_typename(self):
        """Franca 0.9.2, section 5.2.1"""

        with self.assertRaises(ParserException) as context:
            self._parse("""
            package P
            typeCollection TC {
                const double d1 = 123.3d
            }
        """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 4 near 'double'.")

    def test_constants_bad_syntax_hexvalue(self):
        """Franca 0.9.2, section 5.2.1"""

        with self.assertRaises(ParserException) as context:
            self._parse("""
            package P
            typeCollection TC {
                const UInt32 MAX_COUNT = 0xABCDEFUInt32
            }
        """)
        self.assertEqual(str(context.exception),
                         "Syntax error at line 4 near 'UInt32'.")




import os
from collections import OrderedDict
from pyfranca import franca_parser, ast


class ProcessorException(Exception):

    def __init__(self, message):
        super(ProcessorException, self).__init__()
        self.message = message

    def __str__(self):
        return self.message


class Processor(object):
    """
    Franca IDL processor.
    """

    def __init__(self):
        """
        Constructor.
        """
        # Default package paths.
        self.package_paths = []
        self.files = {}
        self.packages = {}

    @staticmethod
    def basename(namespace):
        """
        Extract the type or namespace name from a Franca FQN.
        """
        dot = namespace.rfind(".")
        if dot == -1:
            return namespace
        else:
            return namespace[dot + 1:]

    @staticmethod
    def packagename(namespace):
        """
        Extract the package name from a Franca FQN.
        """
        dot = namespace.rfind(".")
        if dot == -1:
            return None
        else:
            return namespace[0:dot]

    @staticmethod
    def is_fqn(string):
        """
        Defines whether a Franca name is an ID or an FQN.
        """
        return string.count(".") >= 2

    @staticmethod
    def split_fqn(fqn):
        """
        Split a Franca FQN into a tuple - package, namespace, and name.
        """
        parts = fqn.rsplit(".", 2)
        while len(parts) < 3:
            parts.insert(0, None)
        return tuple(parts)

    @staticmethod
    def resolve(namespace, fqn):
        """
        Resolve type references.

        :param namespace: context ast.Namespace object.
        :param fqn: FQN or ID string.
        :return: Dereferenced ast.Type object.
        """
        if not isinstance(namespace, ast.Namespace) or \
                not isinstance(fqn, str):
            raise ValueError("Unexpected input.")
        pkg, ns, name = Processor.split_fqn(fqn)
        if pkg is None:
            # This is an ID
            # Look in the type's namespace
            if name in namespace:
                return namespace[name]
            # Look in other type collections in the type's package
            for typecollection in namespace.package.typecollections.values():
                if name in typecollection:
                    return typecollection[name]
            # Look in imports
            for package_import in namespace.package.imports:
                if package_import.namespace_reference:
                    # Look in namespaces imported in the type's package
                    if name in package_import.namespace_reference:
                        return package_import.namespace_reference[name]
        else:
            # This is an FQN
            if pkg == namespace.package.name:
                # Check in the current package
                if ns in namespace.package.typecollections:
                    if name in namespace.package.typecollections[ns]:
                        return namespace.package.typecollections[ns][name]
            else:
                # Look in typecollections of packages imported in the
                #   type's package using FQNs.
                for package_import in namespace.package.imports:
                    if package_import.namespace == "{}.{}.*".format(pkg, ns):
                        for typecollection in package_import.\
                                package_reference.typecollections.values():
                            if typecollection.name == ns and \
                                    name in typecollection:
                                return typecollection[name]
                        for interface in package_import. \
                                package_reference.interfaces.values():
                            if name in interface:
                                return interface[name]
        # Give up
        raise ProcessorException(
            "Unresolved reference '{}'.".format(fqn))

    @staticmethod
    def resolve_namespace(package, fqn):
        """
        Resolve namespace references.

        :param package: context ast.Package object.
        :param fqn: FQN or ID string.
        :return: Dereferenced ast.Namespace object.
        """
        if not isinstance(package, ast.Package) or not isinstance(fqn, str):
            raise ValueError("Unexpected input.")
        if fqn.count(".") > 0:
            pkg, name = fqn.rsplit(".", 2)
        else:
            pkg, name = (None, fqn)
        if pkg is None:
            # This is an ID
            # Look for other namespaces in the package
            if name in package:
                return package[name]
            # Look in model imports
            for package_import in package.imports:
                if not package_import.namespace:
                    if name in package_import.package_reference:
                        return package_import.package_reference[name]
        else:
            # This is an FQN
            if pkg == package.name:
                # Check in the current package
                if name in package:
                    return package[name]
            else:
                # Look in model imports
                for package_import in package.imports:
                    if not package_import.namespace:
                        if name in package_import.package_reference:
                            return package_import.package_reference[name]
        # Give up
        raise ProcessorException(
            "Unresolved namespace reference '{}'.".format(fqn))

    def _update_complextype_references(self, name):
        """
        Update type references in a complex type.

        :param name: ast.ComplexType object.
        """
        if isinstance(name, ast.Enumeration):
            if name.extends:
                name.reference = self.resolve(name.namespace, name.extends)
                if not isinstance(name.reference, ast.Enumeration):
                    raise ProcessorException(
                        "Invalid enumeration reference '{}'.".format(
                            name.extends))
        elif isinstance(name, ast.Struct):
            for field in name.fields.values():
                self._update_type_references(name.namespace, field.type)
            if name.extends:
                name.reference = self.resolve(name.namespace, name.extends)
                if not isinstance(name.reference, ast.Struct):
                    raise ProcessorException(
                        "Invalid struct reference '{}'.".format(
                            name.extends))
        elif isinstance(name, ast.Array):
            self._update_type_references(name.namespace, name.type)
        elif isinstance(name, ast.Map):
            self._update_type_references(name.namespace, name.key_type)
            self._update_type_references(name.namespace, name.value_type)
        elif isinstance(name, ast.Constant):
            self._update_type_references(name.namespace, name.type)
        else:
            assert False

    def _update_type_references(self, namespace, name):
        """
        Update type references in a type.

        :param namespace: ast.Namespace context.
        :param name: ast.Type object.
        """
        if isinstance(name, ast.Typedef):
            self._update_type_references(name.namespace, name.type)
        elif isinstance(name, ast.PrimitiveType):
            pass
        elif isinstance(name, ast.ComplexType):
            self._update_complextype_references(name)
        elif isinstance(name, ast.Reference):
            if not name.namespace:
                name.namespace = namespace
            if not name.reference:
                resolved_name = self.resolve(namespace, name.name)
                name.reference = resolved_name
        elif isinstance(name, ast.Attribute):
            self._update_type_references(name.namespace, name.type)
        elif isinstance(name, ast.Method):
            for arg in name.in_args.values():
                self._update_type_references(name.namespace, arg.type)
            for arg in name.out_args.values():
                self._update_type_references(name.namespace, arg.type)
            if isinstance(name.errors, OrderedDict):
                pass
            elif isinstance(name.errors, ast.Reference):
                # Errors can be a reference to an enumeration
                self._update_type_references(name.namespace, name.errors)
                if not isinstance(name.errors.reference, ast.Enumeration):
                    raise ProcessorException(
                        "Invalid error reference '{}'.".format(
                            name.errors.name))
            else:
                assert False
        elif isinstance(name, ast.Broadcast):
            for arg in name.out_args.values():
                self._update_type_references(name.namespace, arg.type)
        else:
            assert False

    def _update_namespace_references(self, namespace):
        """
        Update type references in a namespace.

        :param namespace: ast.Namespace object.
        """
        for name in namespace.typedefs.values():
            self._update_type_references(namespace, name)
        for name in namespace.enumerations.values():
            self._update_type_references(namespace, name)
        for name in namespace.structs.values():
            self._update_type_references(namespace, name)
        for name in namespace.arrays.values():
            self._update_type_references(namespace, name)
        for name in namespace.maps.values():
            self._update_type_references(namespace, name)
        for name in namespace.constants.values():
            self._update_type_references(namespace, name)

    def _update_interface_references(self, namespace):
        """
        Update type references in an interface.

        :param namespace: ast.Interface object.
        """
        self._update_namespace_references(namespace)
        for name in namespace.attributes.values():
            self._update_type_references(namespace, name)
        for name in namespace.methods.values():
            self._update_type_references(namespace, name)
        for name in namespace.broadcasts.values():
            self._update_type_references(namespace, name)
        if namespace.extends:
            namespace.reference = self.resolve_namespace(
                namespace.package, namespace.extends)
            if not isinstance(namespace.reference, ast.Interface):
                raise ProcessorException(
                    "Invalid interface reference '{}'.".format(
                        namespace.extends))

    def _update_package_references(self, package):
        """
        Update type references in a package.

        :param package: ast.Package object.
        """
        for package_import in package.imports:
            assert package_import.package_reference is not None
            if package_import.namespace:
                # Namespace import
                package_reference = package_import.package_reference
                if not package_import.namespace.endswith(".*"):
                    raise ProcessorException(
                        "Invalid namespace import {}.".format(
                            package_import.namespace))
                namespace_name = \
                    package_import.namespace[len(package_reference.name) + 1:-2]
                # Update namespace reference
                if namespace_name in package_reference:
                    namespace = package_reference[namespace_name]
                    package_import.namespace_reference = namespace
                else:
                    raise ProcessorException(
                        "Namespace '{}' not found.".format(
                            package_import.namespace))
            else:
                # Model import
                assert package_import.namespace_reference is None
        for namespace in package.typecollections:
            self._update_namespace_references(
                package.typecollections[namespace])
        for namespace in package.interfaces:
            self._update_interface_references(
                package.interfaces[namespace])

    def import_package(self, fspec, package, references=None):
        """
        Import an ast.Package into the processor.

        :param fspec: File specification of the package.
        :param package: ast.Package object.
        :param references: A list of package references.
        """
        if not isinstance(package, ast.Package):
            ValueError("Expected ast.Package as input.")
        if not references:
            references = []
        # Check whether package is already imported
        if package.name in self.packages:
            if fspec not in self.packages[package.name].files:
                # Merge the new package into the already existing one.
                self.packages[package.name] += package
                # Register the package file in the processor.
                self.files[fspec] = self.packages[package.name]
                package = self.packages[package.name]
            else:
                return
        else:
            # Register the package in the processor.
            self.packages[package.name] = package
            # Register the package file in the processor.
            self.files[fspec] = package
        # Process package imports
        fspec_path = os.path.abspath(fspec)
        fspec_dir = os.path.dirname(fspec_path)
        for package_import in package.imports:
            imported_package = self.import_file(
                package_import.file, references + [package.name], fspec_dir)
            # Update import reference
            package_import.package_reference = imported_package
        # Update type references
        self._update_package_references(package)

    def import_string(self, fspec, fidl, references=None):
        """
        Parse an FIDL string and import it into the processor as package.

        :param fspec: File specification of the package.
        :param fidl: FIDL string.
        :param references: A list of package references.
        :return: The parsed ast.Package.
        """
        # Parse the string.
        parser = franca_parser.Parser()
        package = parser.parse(fidl)
        package.files = [fspec]
        # Import the package in the processor.
        self.import_package(fspec, package, references)
        return package

    def import_file(self, fspec, references=None, package_path=None):
        """
        Parse an FIDL file and import it into the processor as package.

        :param fspec: File specification.
        :param references: A list of package references.
        :param package_path: Additional model path to search for imports.
        :return: The parsed ast.Package.
        """
        if fspec in self.files:
            # File already loaded.
            return self.files[fspec]
        if not os.path.exists(fspec):
            if os.path.isabs(fspec):
                # Absolute specification
                raise ProcessorException(
                    "Model '{}' not found.".format(fspec))
            else:
                # Relative specification.
                package_paths = self.package_paths[:]
                if package_path:
                    package_paths.insert(0, package_path)
                # Check in the package path list.
                for path in package_paths:
                    temp_fspec = os.path.join(path, fspec)
                    if os.path.exists(temp_fspec):
                        fspec = temp_fspec
                        break
                else:
                    raise ProcessorException(
                        "Model '{}' not found.".format(fspec))
        # Parse the file.
        parser = franca_parser.Parser()
        package = parser.parse_file(fspec)
        # Import the package in the processor.
        self.import_package(fspec, package, references)
        return package



#!/usr/bin/env python

import argparse
from pyfranca import Processor, LexerException, ParserException, \
    ProcessorException


def parse_command_line():
    parser = argparse.ArgumentParser(
        description="Behavioral cloning model trainer.")
    parser.add_argument(
        "fidl", nargs="+",
        help="Input FIDL file.")
    parser.add_argument(
        "-I", "--import", dest="import_dirs", metavar="import_dir",
        action="append", help="Model import directories.")
    args = parser.parse_args()
    return args


def main():
    args = parse_command_line()

    processor = Processor()
    if args.import_dirs:
        processor.package_paths.extend(args.import_dirs)

    try:
        for fidl in args.fidl:
            processor.import_file(fidl)
    except (LexerException, ParserException, ProcessorException) as e:
        print("ERROR: {}".format(e))
        exit(1)

    print("Valid Franca model.")


if __name__ == "__main__":
    main()



"""
Pyfranca processor tests.
"""

import unittest
import os
import errno
import shutil

from pyfranca import ProcessorException, Processor, ast


class BaseTestCase(unittest.TestCase):

    def setUp(self):
        # Create temporary file directory.
        tmp_dir = self.get_spec()
        try:
            os.makedirs(tmp_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            # Delete and try again
            shutil.rmtree(tmp_dir)
            os.makedirs(tmp_dir)
        # Create FIDL processor.
        self.processor = Processor()
        self.processor.package_paths.append(tmp_dir)

    def tearDown(self):
        self.processor = None
        # Remove temporary file directory.
        tmp_dir = self.get_spec()
        shutil.rmtree(tmp_dir)

    @staticmethod
    def get_spec(basedir="tmp", filename=None):
        """
        Get absolute specification a directory or a file under pyfranca/tests/fidl/ .
        :param basedir: Target subdirectory.
        :param filename: File name or None to get a directory specification
        :return: Absolute specification.
        """
        script_dir = os.path.dirname(os.path.realpath(__file__))
        spec = os.path.join(script_dir, "fidl", basedir)
        if filename:
            spec = os.path.join(spec, filename)
        return spec

    def tmp_fidl(self, filename, content):
        fspec = self.get_spec(filename=filename)
        with open(fspec, "w") as f:
            f.write(content)
        return fspec

    def import_tmp_fidl(self, filename, content):
        fspec = self.tmp_fidl(filename, content)
        self.processor.import_file(fspec)


class TestFQNs(BaseTestCase):
    """Test FQN parsing methods."""

    def test_basename(self):
        self.assertEqual(Processor.basename("P.I.A"), "A")
        self.assertEqual(Processor.basename("A"), "A")

    def test_packagename(self):
        self.assertEqual(Processor.packagename("P.I.A"), "P.I")
        self.assertIsNone(Processor.packagename("A"))

    def test_is_fqn(self):
        self.assertTrue(Processor.is_fqn("P.P.I.A"))
        self.assertTrue(Processor.is_fqn("P.I.A"))
        self.assertFalse(Processor.is_fqn("I.A"))
        self.assertFalse(Processor.is_fqn("A"))

    def test_split_fqn(self):
        self.assertEqual(Processor.split_fqn("P.P.I.A"), ("P.P", "I", "A"))
        self.assertEqual(Processor.split_fqn("P.I.A"), ("P", "I", "A"))
        self.assertEqual(Processor.split_fqn("I.A"), (None, "I", "A"))
        self.assertEqual(Processor.split_fqn("A"), (None, None, "A"))


class TestImports(BaseTestCase):
    """Test import related features."""

    def test_basic(self):
        self.processor.import_string("test.fidl", """
            package P
        """)
        self.processor.import_string("test2.fidl", """
            package P2
        """)
        # Verify file access
        self.assertEqual(len(self.processor.files), 2)
        p = self.processor.files["test.fidl"]
        self.assertEqual(p.name, "P")
        self.assertEqual(p.files, ["test.fidl"])
        p2 = self.processor.files["test2.fidl"]
        self.assertEqual(p2.name, "P2")
        self.assertEqual(p2.files, ["test2.fidl"])
        # Verify package access
        self.assertEqual(len(self.processor.packages), 2)
        p = self.processor.packages["P"]
        self.assertEqual(p.name, "P")
        p2 = self.processor.packages["P2"]
        self.assertEqual(p2.name, "P2")

    def test_import_nonexistent_model(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                import model "nosuch.fidl"
            """)
        self.assertEqual(str(context.exception),
                         "Model 'nosuch.fidl' not found.")

    def test_import_nonexistent_model2(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                import P.Nonexistent.* from "nosuch.fidl"
            """)
        self.assertEqual(str(context.exception),
                         "Model 'nosuch.fidl' not found.")

    def test_import_nonexistent_namespace(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
            """)
            self.processor.import_string("test2.fidl", """
                package P2
                import P.Nonexistent.* from "test.fidl"
            """)
        self.assertEqual(str(context.exception),
                         "Namespace 'P.Nonexistent.*' not found.")

    def test_circular_dependency(self):
        fspec = self.tmp_fidl("test.fidl", """
            package P
            import model "test2.fidl"
        """)
        self.tmp_fidl("test2.fidl", """
            package P2
            import model "test.fidl"
        """)
        self.processor.import_file(fspec)
        # FIXME: What is the correct behavior?


class TestPackagesInMultipleFiles(BaseTestCase):
    """Support for packages in multiple files"""

    def test_package_in_multiple_files(self):
        self.processor.import_string("test.fidl", """
            package P
        """)
        self.processor.import_string("test2.fidl", """
            package P
        """)
        p = self.processor.packages["P"]
        self.assertEqual(p.files, ["test.fidl", "test2.fidl"])

    def test_package_in_multiple_files_reuse(self):
        self.processor.import_string("test.fidl", """
            package P
            typeCollection TC {
                typedef A is Int32
            }
        """)
        self.processor.import_string("test2.fidl", """
            package P
            import P.TC.* from "test.fidl"
            interface I {
                typedef B is A
            }
        """)
        p = self.processor.packages["P"]
        self.assertEqual(p.files, ["test.fidl", "test2.fidl"])
        a = p.typecollections["TC"].typedefs["A"]
        self.assertEqual(a.namespace.package, p)
        self.assertTrue(isinstance(a.type, ast.Int32))
        b = p.interfaces["I"].typedefs["B"]
        self.assertEqual(b.namespace.package, p)
        self.assertTrue(isinstance(b.type, ast.Reference))
        self.assertEqual(b.type.name, "A")
        self.assertEqual(b.type.reference, a)


class TestReferences(BaseTestCase):
    """Test type references."""

    def test_reference_to_same_namespace(self):
        self.processor.import_string("test.fidl", """
            package P
            typeCollection TC {
                typedef A is Int32
                typedef B is A
            }
        """)
        a = self.processor.packages["P"].typecollections["TC"].typedefs["A"]
        self.assertTrue(isinstance(a.type, ast.Int32))
        b = self.processor.packages["P"].typecollections["TC"].typedefs["B"]
        self.assertTrue(isinstance(b.type, ast.Reference))
        self.assertEqual(b.type.name, "A")
        self.assertEqual(b.type.reference, a)

    def test_fqn_reference(self):
        self.processor.import_string("test.fidl", """
            package P
            typeCollection TC {
                typedef A is Int32
                typedef B is P.TC.A
            }
        """)
        a = self.processor.packages["P"].typecollections["TC"].typedefs["A"]
        self.assertTrue(isinstance(a.type, ast.Int32))
        b = self.processor.packages["P"].typecollections["TC"].typedefs["B"]
        self.assertTrue(isinstance(b.type, ast.Reference))
        self.assertEqual(b.type.name, "P.TC.A")
        self.assertEqual(b.type.reference, a)

    def test_reference_to_different_namespace(self):
        self.processor.import_string("test.fidl", """
            package P
            typeCollection TC {
                typedef A is Int32
            }
            interface I {
                typedef B is A
            }
        """)
        a = self.processor.packages["P"].typecollections["TC"].typedefs["A"]
        self.assertTrue(isinstance(a.type, ast.Int32))
        b = self.processor.packages["P"].interfaces["I"].typedefs["B"]
        self.assertTrue(isinstance(b.type, ast.Reference))
        self.assertEqual(b.type.name, "A")
        self.assertEqual(b.type.reference, a)

    def test_reference_to_different_model(self):
        self.processor.import_string("test.fidl", """
            package P
            typeCollection TC {
                typedef A is Int32
            }
        """)
        self.processor.import_string("test2.fidl", """
            package P2
            import P.TC.* from "test.fidl"
            interface I {
                typedef B is A
            }
        """)
        a = self.processor.packages["P"].typecollections["TC"].typedefs["A"]
        self.assertTrue(isinstance(a.type, ast.Int32))
        b = self.processor.packages["P2"].interfaces["I"].typedefs["B"]
        self.assertTrue(isinstance(b.type, ast.Reference))
        self.assertEqual(b.type.name, "A")
        self.assertEqual(b.type.reference, a)

    @unittest.skip("Currently not checked.")
    def test_circular_reference(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                typeCollection TC {
                    typedef A is B
                    typedef B is A
                }
            """)
        self.assertEqual(str(context.exception),
                         "Circular reference 'B'.")

    def test_reference_priority(self):
        self.processor.import_string("test.fidl", """
            package P
            typeCollection TC {
                typedef A is Int32
            }
        """)
        self.processor.import_string("test2.fidl", """
            package P2
            import P.TC.* from "test.fidl"
            typeCollection TC {
                typedef A is UInt32
            }
            interface I {
                typedef B is A
                typedef B2 is P.TC.A
            }
        """)
        a = self.processor.packages["P"].typecollections["TC"].typedefs["A"]
        self.assertTrue(isinstance(a.type, ast.Int32))
        a2 = self.processor.packages["P2"].typecollections["TC"].typedefs["A"]
        self.assertTrue(isinstance(a2.type, ast.UInt32))
        b = self.processor.packages["P2"].interfaces["I"].typedefs["B"]
        self.assertTrue(isinstance(b.type, ast.Reference))
        self.assertEqual(b.type.name, "A")
        self.assertEqual(b.type.reference, a2)
        b2 = self.processor.packages["P2"].interfaces["I"].typedefs["B2"]
        self.assertTrue(isinstance(b2.type, ast.Reference))
        self.assertEqual(b2.type.name, "P.TC.A")
        self.assertEqual(b2.type.reference, a)

    def test_interface_visibility(self):
        self.processor.import_string("test.fidl", """
            package P
            typeCollection TC {
                typedef A is Int32
            }
        """)
        self.processor.import_string("test2.fidl", """
            package P2
            import P.TC.* from "test.fidl"
            interface I {
                typedef A is UInt32
            }
            typeCollection TC {
                typedef B is A
            }
        """)
        a = self.processor.packages["P"].typecollections["TC"].typedefs["A"]
        self.assertTrue(isinstance(a.type, ast.Int32))
        a2 = self.processor.packages["P2"].interfaces["I"].typedefs["A"]
        self.assertTrue(isinstance(a2.type, ast.UInt32))
        b = self.processor.packages["P2"].typecollections["TC"].typedefs["B"]
        self.assertTrue(isinstance(b.type, ast.Reference))
        self.assertEqual(b.type.name, "A")
        self.assertEqual(b.type.reference, a)

    def test_unresolved_reference_in_typedef(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                typeCollection TC {
                    typedef TD is Unknown
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_struct(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                typeCollection TC {
                    struct S {
                        Unknown f
                    }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_array(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                typeCollection TC {
                    array A of Unknown
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_map(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                typeCollection TC {
                    map M { Unknown to Int32 }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_map2(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                typeCollection TC {
                    map M { Int32 to Unknown }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_attribute(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                interface I {
                    attribute Unknown A
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_method(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                interface I {
                    method M { in { Unknown a } }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_method2(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                interface I {
                    method M { out { Unknown a } }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_method3(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                interface I {
                    method M { error Unknown }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_unresolved_reference_in_broadcast(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                interface I {
                    broadcast B { out { Unknown a } }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved reference 'Unknown'.")

    def test_method_error_reference(self):
        self.processor.import_string("test.fidl", """
            package P
            interface I {
                enumeration E { A B C }
                method M { error E }
            }
        """)
        e = self.processor.packages["P"].interfaces["I"].enumerations["E"]
        m = self.processor.packages["P"].interfaces["I"].methods["M"]
        me = m.errors
        self.assertTrue(isinstance(me, ast.Reference))
        self.assertEqual(me.reference, e)

    def test_invalid_method_error_reference(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                interface I {
                    typedef E is String
                    method M { error E }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Invalid error reference 'E'.")

    def test_enumeration_extension(self):
        self.processor.import_string("test.fidl", """
            package P
            interface I {
                enumeration E { A B C }
                enumeration E2 extends E { D E F }
            }
        """)
        e = self.processor.packages["P"].interfaces["I"].enumerations["E"]
        e2 = self.processor.packages["P"].interfaces["I"].enumerations["E2"]
        self.assertEqual(e2.reference, e)

    def test_invalid_enumeration_extension(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                interface I {
                    typedef E is String
                    enumeration E2 extends E { D E F }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Invalid enumeration reference 'E'.")

    def test_struct_extension(self):
        self.processor.import_string("test.fidl", """
            package P
            interface I {
                struct S { Int32 a }
                struct S2 extends S { Int32 b }
            }
        """)
        s = self.processor.packages["P"].interfaces["I"].structs["S"]
        s2 = self.processor.packages["P"].interfaces["I"].structs["S2"]
        self.assertEqual(s2.reference, s)

    def test_invalid_struct_extension(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                interface I {
                    typedef S is String
                    struct S2 extends S { Int32 b }
                }
            """)
        self.assertEqual(str(context.exception),
                         "Invalid struct reference 'S'.")

    def test_interface_extension(self):
        self.processor.import_string("test.fidl", """
            package P
            interface I { }
            interface I2 extends I { }
            interface I3 extends P.I { }
        """)
        i = self.processor.packages["P"].interfaces["I"]
        i2 = self.processor.packages["P"].interfaces["I2"]
        self.assertEqual(i2.reference, i)
        i3 = self.processor.packages["P"].interfaces["I3"]
        self.assertEqual(i3.reference, i)

    def test_interface_extension2(self):
        self.processor.import_string("test.fidl", """
            package P
            interface I { }
        """)
        self.processor.import_string("test2.fidl", """
            package P2
            import model "test.fidl"
            interface I2 extends I { }
            interface I3 extends P.I { }
        """)
        i = self.processor.packages["P"].interfaces["I"]
        i2 = self.processor.packages["P2"].interfaces["I2"]
        self.assertEqual(i2.reference, i)
        i3 = self.processor.packages["P2"].interfaces["I3"]
        self.assertEqual(i3.reference, i)

    def test_invalid_interface_extension(self):
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_string("test.fidl", """
                package P
                typeCollection TC { typedef I is Int32 }
                interface I2 extends I { }
            """)
        self.assertEqual(str(context.exception),
                         "Unresolved namespace reference 'I'.")

    def test_anonymous_array_references(self):
        self.processor.import_string("test.fidl", """
            package P
            typeCollection TC {
                typedef TD is Int32
                typedef TDA is TD[]
                array ATDA of TD[]
                struct S { TD[] tda }
                map M { TD[] to TD[] }
            }
            interface I {
                attribute TD[] A
                method M { in { TD[] tda } out { TD[] tda } }
                broadcast B { out { TD[] tda } }
            }
        """)
        tc = self.processor.packages["P"].typecollections["TC"]
        td = tc.typedefs["TD"]
        tda = tc.typedefs["TDA"]
        self.assertEqual(tda.type.type.reference, td)
        atda = tc.arrays["ATDA"]
        self.assertEqual(atda.type.type.reference, td)
        s = tc.structs["S"]
        self.assertEqual(s.fields["tda"].type.type.reference, td)
        m = tc.maps["M"]
        self.assertEqual(m.key_type.type.reference, td)
        self.assertEqual(m.value_type.type.reference, td)
        i = self.processor.packages["P"].interfaces["I"]
        a = i.attributes["A"]
        self.assertEqual(a.type.type.reference, td)
        m = i.methods["M"]
        self.assertEqual(m.in_args["tda"].type.type.reference, td)
        self.assertEqual(m.out_args["tda"].type.type.reference, td)
        b = i.broadcasts["B"]
        self.assertEqual(b.out_args["tda"].type.type.reference, td)

    def test_import_missing_files(self):
        # P.fidl references definitions.fidl but it is not in the package path.
        with self.assertRaises(ProcessorException) as context:
            self.processor.import_file(
                os.path.join("pyfranca", "tests", "fidl", "idl", "P.fidl"))
            self.processor.import_file(
                os.path.join("pyfranca", "tests", "fidl", "idl2", "P2.fidl"))
        self.assertEqual(str(context.exception),
                         "Model 'definitions.fidl' not found.")

    def test_import_multiple_files(self):
        script_dir = os.path.dirname(os.path.realpath(__file__))
        fidl_dir = os.path.join(script_dir, "fidl", "idl")
        self.processor.package_paths.append(fidl_dir)
        self.processor.import_file(
            os.path.join("P.fidl"))
        self.processor.import_file(
            os.path.join("pyfranca", "tests", "fidl", "idl2", "P2.fidl"))

    def test_import_file_chain(self):
        fspec = self.tmp_fidl("P.fidl", """
            package P
            import P.Type1.* from "Type1.fidl"
            import P.Type2.* from "Type2.fidl"
            interface I {
                version { major 1 minor 0 }
                method getData {
                    out {
                          MyStructType1 outVal
                       }
                    in {
                          MyStructType2 inVal
                       }
                }
            }
        """)
        self.tmp_fidl("Type1.fidl", """
            package P
            import P.Common.* from "common.fidl"
            typeCollection Type1 {
                version { major 1 minor 0 }
                struct MyStructType1 {
                    UInt8 val1
                    MyEnum val2
                    String  msg
                }
            }
        """)
        self.tmp_fidl("Type2.fidl", """
            package P
            import P.Common.* from "common.fidl"
            typeCollection Type2 {
                version { major 1 minor 0 }
                struct MyStructType2 {
                    UInt8 val1
                    MyEnum val2
                    String  msg
                }
            }
        """)
        self.tmp_fidl("common.fidl", """
            package P
            typeCollection Common {
                version { major 1 minor 0 }
                enumeration MyEnum {
                    abc
                    def
                    ghi
                    jkm
                }
            }
        """)
        self.processor.import_file(fspec)



"""
Pyfranca package.
"""

from pyfranca.franca_lexer import LexerException, Lexer
from pyfranca.franca_parser import ParserException, Parser
from pyfranca.franca_processor import ProcessorException, Processor


__version__ = "0.4.0"



#!/usr/bin/env python
"""
Pyfranca setup script.
"""

import os
import re
from setuptools import setup, find_packages


def read_package_variable(key):
    fspec = os.path.join("pyfranca", "__init__.py")
    with open(fspec) as f:
        for line in f:
            m = re.match(r"(\S+)\s*=\s*[\"']?(.+?)[\"']?\s*$", line)
            if m and key == m.group(1):
                return m.group(2)
    return None


version = read_package_variable("__version__")

setup(
    name="pyfranca",
    packages=find_packages(),
    version=version,
    description="Python parser and tools for working with the Franca "
                "interface definition language.",
    author="Kaloyan Tenchov",
    author_email="zayfod@gmail.com",
    url="http://github.com/zayfod/pyfranca",
    download_url="https://github.com/zayfod/pyfranca/archive/{}.zip".format(
        version),
    license="MIT",
    platforms="Python 2.7 or 3.4 .",
    keywords=["franca", "franca-idl", "idl", "fidl", "parsing"],
    install_requires=["ply"],
    setup_requires=[
        'setuptools_pep8',
        'sphinx',
    ],
    test_suite="pyfranca.tests.get_suite",
    scripts=[
        "tools/fidl_dump.py",
        "tools/fidl_validator.py",
    ],
)
